#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "softcris.h"
#include "hdfi.h"
#include "crisExpEvt.h"
#include "ce_crisEvt.h"
#include "level1.h"
#include "getEvents_c.h"
#include "ce2.h"

    /* debug */
#define VERBOSE 0
#define VERBOSE_SOFT_EXCEPTIONS_FOR_REJECTED_EVENTS 0
#define DEBUG_HEADER_INFO_GENERATION 0

    /* assume there won't be more than 500 events in an instrument cycle */
#define MAX_EVENTS_PER_FRAME 500

#define COMPRESSED_ELEMENTS 256
#define NUMBER_OF_DESIRED_TRAJECTORIES 1

u_char compressNpix(long npix);
long uncompNpix(u_char compNpix);
double interpol(u_char *compressed, long *uncompressed, long npix);
u_char compressInt(long intensity);
void processSOFTException(int errorCode);
 
struct ce2 *recompressFramesEvents(struct ce_crisExpEvt *pExpEvt,
		struct sc_evt_params *PTR_sc_evt_params_structs,
		long max_layer_intensities[][6], int hop_params[],
		float qual_factors[], int eventCount);
void writeFramesRecompressedEvents(int fd, struct ce2 *pRecompressedEvents,
								int eventCount);
int get_command_line_args(int argc, char *argv[], char user[],
				char output_directory[], char institution[]);
void print_usage();
void get_version_information_from_version_files(char directory[],
			char *version_file_names[],
			struct output_file_header_info *PTR_header_info);
void write_ce2_file_header(int fd,
			struct output_file_header_info *PTR_header_info);

 
int main(int argc, char *argv[]) {
    int day, startDay1996, endDay1996;

    char output_directory[80] = ".";	/* default is cwd */
    struct output_file_header_info header_info =
			{ 0x30313233, "", "", "", "", "", "", "", "", "" };

    argc = get_command_line_args(argc, argv, header_info.user,
				header_info.institution, output_directory);

/*
    {
	int i;
	printf("output_directory = %s\n", output_directory);
	printf("user = %s\n", header_info.user);
	printf("institution = %s\n", header_info.institution);

	for (i = 0; i < argc; i++) {
	    printf("%d: %s\n", i, argv[i]);
	}
    }
*/

    if (argc != 3) {
	fprintf(stderr, "\nIncorrect number of arguments.\n\n");
	print_usage();
	exit(0);
    }
	/* institution must match one of (caltech|washu|gsfc) strings */
    if (		   strcmp(header_info.institution, "caltech")
			&& strcmp(header_info.institution, "washu")
			&& strcmp(header_info.institution, "gsfc")   ) {
	fprintf(stderr, "\nInstitution = \"%s\" invalid.\n\n",
						header_info.institution);
	print_usage();
	exit(0);
    }
	/* does output directory exist? */
    if (strcmp(output_directory, ".")) {
	if (!directoryExists(output_directory)) {
	    fprintf(stderr, "\noutput directory = %s does not exist.\n\n",
							output_directory);
	    print_usage();
	    exit(0);
	}
    }
	/* do header info files exist? */
    {
	char *version_file_names[] = { "VERSION_level1", "VERSION_level1_1",
			"VERSION_mk_ce2", "VERSION_softlib", "DATE_compiled" };
	char base_directory[] = "ce2_file_header";
	char directory[80] = "";
	char file_name[80] = "";
	int i;

	sprintf(directory, "%s/%s", base_directory, header_info.institution);

	for (i = 0; i < 5; i++) {
	    sprintf(file_name, "%s/%s", directory, version_file_names[i]);
	    if (!fileExists(file_name)) {
		fprintf(stderr, "\nVersion info file %s missing.  Please see\n",
								file_name);
		fprintf(stderr, "documentation in %s/README.\n\n",
							base_directory);
		exit(0);
	    }
	}
	get_version_information_from_version_files(directory,
					version_file_names, &header_info);
    }
	/* can we determine the user creating the data files? */
    if (!strcmp(header_info.user, "")) {
	    /* user was not in the arguments, try USER environment var */
    	if (getenv("USER") != NULL) {
	    sprintf(header_info.user, "%s", getenv("USER"));
	}

    	if (!strcmp(header_info.user, "")) {
	    fprintf(stderr, "\nA username for the creator of the data\n");
	    fprintf(stderr, "files must be specified either by a USER\n");
	    fprintf(stderr, "environment variable or as a command line\n");
	    fprintf(stderr, "variable.  This information is used so\n");
	    fprintf(stderr,"that a potential scapegoat may be identified\n");
	    fprintf(stderr, "in the header of the data file.\n\n");
	    print_usage();
	    exit(0);
	}
    }
	/* we also want some clue as to the version of the archive being used */
    sprintf(header_info.data_archive_version, "%s",
			getenv("L1_CRIS_DATA_BASE_DIRECTORY"));

	/* get the runtime of the creation of this file */
    {
	time_t time_value;
	struct tm *PTR_time_struct;

	time_value = time(NULL);
	PTR_time_struct = localtime(&time_value);
	strftime(header_info.run_date, 80, "%a %b %d %H:%M:%S %Z %Y",
							PTR_time_struct);
    }

    if (DEBUG_HEADER_INFO_GENERATION) {
	printf("institution = %s\n", header_info.institution);
	printf("user = %s\n", header_info.user);
	printf("mk_ce2_version = %s\n", header_info.mk_ce2_version);
	printf("softlib_version = %s\n", header_info.softlib_version);
	printf("level1_version = %s\n", header_info.level1_version);
	printf("level1_1_version = %s\n", header_info.level1_1_version);
	printf("data_archive_version = %s\n", header_info.data_archive_version);
	exit(0);
    }

    startDay1996 = atoi(argv[1]);
    endDay1996 = atoi(argv[2]);

	/* set trajectory method in SOFT lib */
	/* brightest-each-layer -> 0, fiber-fit -> 1-4, multi-traj -> 5-8 */
    sc_traj_select_method(8);

	/* Begin looping through event cycles on a given day. */
    for (day = startDay1996; day <= endDay1996; day++) {
	int fd;
	char filePath[256];
	uint32 *pRequestTimes;

		/* check existance of inputArchiveFile */
	{
	    static char pathBuffer[256];
	    static const char instrumentID = 'c';
	    static const char dataFileSuffix[] = "evt";
	    struct tm *pTimeStruct;
	    
	    pTimeStruct = getDateFromDay1996(day);
	    generateDataFilePath(pathBuffer, pTimeStruct, instrumentID,
								dataFileSuffix);

	    if (!fileExists(pathBuffer)) {	/* go to next day */
		fprintf(stderr, "Warning: Input archive file %s ", pathBuffer);
		fprintf(stderr, "does not exist.\n");
		fprintf(stderr, "\tContinuing to next day.\n\n");
		continue;
	    }
	}

	/* sprintf(filePath, "%s/events.%4.4u", output_directory, day); */
	sprintf(filePath, "%s/c_%4.4u.ce2", output_directory, day);
	fd = openOrCreateFileToReadWrite(filePath);
	write_ce2_file_header(fd, &header_info);

	pRequestTimes = getRequestTimesForDaysEvents_c(day, 1);

	    /* used to collect all the information needed for filling the
	     * ce1 fields in the array of ce_crisExpEvt structs, but now
	     * for the new ce2 struct, instead of playing around with the
	     * ce_crisExpEvt struct, we just made a couple more arrays
	     * to handle storing added information until the ce2 structs
	     * are created and written to file... (e.g., hop_params array)
	     */
	while (*pRequestTimes != 0) {
	    static struct ce_crisEvt *pL1Event;
	    static struct ce_crisExpEvt pExpEvt[MAX_EVENTS_PER_FRAME];
	    static struct sc_evt_params
			sc_evt_params_structs[MAX_EVENTS_PER_FRAME];
	    static long max_layer_intensities[MAX_EVENTS_PER_FRAME][6];
	    static int hop_params[MAX_EVENTS_PER_FRAME];
	    static float qual_factors[MAX_EVENTS_PER_FRAME];

	    u_char goodTrajectory;
	    int iEvt = 0;

	    pL1Event = getEvents_c(*pRequestTimes);

	    while (pL1Event->secs1996 != 0) {
		struct sc_Coords_Traj
		    coordsAndTrajectoriesInfo[NUMBER_OF_DESIRED_TRAJECTORIES];

		int errorCode, numberOfTrajectories;
		static struct sc_fmapElem *pFiberMapElement;
		int result, i;

		if (iEvt >= MAX_EVENTS_PER_FRAME) {
		    printf("Error: Exceeding maximum of %d events/frame.\n",
							MAX_EVENTS_PER_FRAME);
		    printf("Exitting.");
		    exit(1);
		}
	    
		goodTrajectory = 0;   /* assume bad trajectory at start */
		for (i = 0; i < CE_CRIS_MAXBLOBS; i++) { /* initialize */
		    pExpEvt[iEvt].blobpos[i] = -100.0; /* needs to be unique */
		    pExpEvt[iEvt].layer[i] = (u_char) 32;
		}
		pExpEvt[iEvt].x1 = 0.0;
		pExpEvt[iEvt].x2 = 0.0;
		pExpEvt[iEvt].x3 = 0.0;
		pExpEvt[iEvt].y1 = 0.0;
		pExpEvt[iEvt].y2 = 0.0;
		pExpEvt[iEvt].y3 = 0.0;
		pExpEvt[iEvt].ix1 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].ix2 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].ix3 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy1 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy2 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy3 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].nx1 = (u_char) 255; /* 255 marks invalid value */
		pExpEvt[iEvt].nx2 = (u_char) 255;
		pExpEvt[iEvt].nx3 = (u_char) 255;
		pExpEvt[iEvt].ny1 = (u_char) 255;
		pExpEvt[iEvt].ny2 = (u_char) 255;
		pExpEvt[iEvt].ny3 = (u_char) 255;
		pExpEvt[iEvt].goodtraj = (u_char) 0;

		for (i = 0; i < ((int) pL1Event->nblobs); i++) {
		    float xreal, yreal;
			/* layer ID and middle fiber ID for each layer */
		    static int layerAndMidFiberIDs[6] =
					{ 1940, 640, 4503, 3248, 6828, 5686 };
		    int layer;

		    pFiberMapElement = sc_centr_to_fiber(SC_CAM2,
					(double) pL1Event->hodo[i].x,
					(double) pL1Event->hodo[i].y);
			/* unmapped coords? */
		    if ((pFiberMapElement == NULL) ||
					(pFiberMapElement->layer == -1)) {
			pExpEvt[iEvt].blobpos[i] = -10.0;
			pExpEvt[iEvt].layer[i] = (u_char) 255;
			continue;
		    }

		    layer = 3 - ((int) (pFiberMapElement->layer / 2));

		    if ((pFiberMapElement->layer % 2) == 0) {
			    /* this is a y-layer fiber */
			result = sc_fiber_to_real(SC_CAM2, layer,
				layerAndMidFiberIDs[pFiberMapElement->layer],
				pFiberMapElement->fiber, &xreal, &yreal);
			if (result == 0) { /* all OK */
			    pExpEvt[iEvt].blobpos[i] = (float) yreal;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
			else {
			    pExpEvt[iEvt].blobpos[i] = -1.0;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
		    }
		    else {			/* this is an x-layer fiber */
			result = sc_fiber_to_real(SC_CAM2, layer,
				pFiberMapElement->fiber,
				layerAndMidFiberIDs[pFiberMapElement->layer],
				&xreal, &yreal);
			if (result == 0) {
			    pExpEvt[iEvt].blobpos[i] = (float) xreal;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
			else {
			    pExpEvt[iEvt].blobpos[i] = -1.0;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
		    } 
		}
		numberOfTrajectories = 1;
		result = sc_cris_traj(pL1Event, coordsAndTrajectoriesInfo,
							&numberOfTrajectories);
		errorCode = sc_get_err_code();

		    /* get more trajectory info.  important to call
		     * this routine just after sc_cris_traj.
	 	     */
		sc_get_evt_params(&sc_evt_params_structs[iEvt]);

		if (result == 0) {		/* found a good trajectory */

			/* get trajectory quality info. */
		    hop_params[iEvt] = coordsAndTrajectoriesInfo[0].Q.hop_param;
		    qual_factors[iEvt] =
				coordsAndTrajectoriesInfo[0].Q.qual_factor;

			/* get max intensities in layers.  important to call
			 * this routine just after sc_cris_traj.
		 	 */
		    {
			long *I;
			int i;

			I = sc_layer_maxI();
			    /* I is returned in the order y3, x3, ... x1 */
			    /* we want to order it x1, y1, ... y3 */
			for (i = 0; i < 6; i++) {
			    max_layer_intensities[iEvt][i] = I[5 - i];
			}
		    }

			/* this flag is being set to 2 instead of 1 in order */
			/* that the generated data files agree with an older */
			/* version of this routine. */
		    goodTrajectory = 2;

			/* top layer to bottom layer in SOFT */
		    pExpEvt[iEvt].x1 = coordsAndTrajectoriesInfo[0].T.layer1x;
		    pExpEvt[iEvt].y1 = coordsAndTrajectoriesInfo[0].T.layer1y;
		    pExpEvt[iEvt].x2 = coordsAndTrajectoriesInfo[0].T.layer2x;
		    pExpEvt[iEvt].y2 = coordsAndTrajectoriesInfo[0].T.layer2y;
		    pExpEvt[iEvt].x3 = coordsAndTrajectoriesInfo[0].T.layer3x;
		    pExpEvt[iEvt].y3 = coordsAndTrajectoriesInfo[0].T.layer3y;

			/* determine th blob IDs used in the trajectory. */
		    pExpEvt[iEvt].ny3 = coordsAndTrajectoriesInfo[0].blobid[0];
		    pExpEvt[iEvt].nx3 = coordsAndTrajectoriesInfo[0].blobid[1];
		    pExpEvt[iEvt].ny2 = coordsAndTrajectoriesInfo[0].blobid[2];
		    pExpEvt[iEvt].nx2 = coordsAndTrajectoriesInfo[0].blobid[3];
		    pExpEvt[iEvt].ny1 = coordsAndTrajectoriesInfo[0].blobid[4];
		    pExpEvt[iEvt].nx1 = coordsAndTrajectoriesInfo[0].blobid[5];
		}
		else processSOFTException(errorCode);

		pExpEvt[iEvt].EvtData = *pL1Event;
		pExpEvt[iEvt].goodtraj = goodTrajectory;

		pL1Event++;
		iEvt++; 
	    }

	    /* printf("frame= %d, eventCount = %d\n", loop, iEvt); */

		/* recompress events and write them to file */
	    {
		struct ce2 *pRecompressedEvents, *pRecompressedEventsStart;

		    /* pass all info needed to create all the ce2 structs
		     * in this frame.
		     */
		pRecompressedEventsStart = recompressFramesEvents(pExpEvt,
			sc_evt_params_structs, max_layer_intensities,
			hop_params, qual_factors, iEvt);
		pRecompressedEvents = pRecompressedEventsStart;
		writeFramesRecompressedEvents(fd, pRecompressedEvents, iEvt);
		free(pRecompressedEventsStart);
	    }

	    pRequestTimes++;
	}
	close(fd);
    }
    return 0;
}

void writeFramesRecompressedEvents(int fd, struct ce2 *pRecompressedEvents,
								int count) {
    long bytesLeftToWrite;
    int DEVICE_BLOCK_SIZE = 8192;
    char *pChar;

    bytesLeftToWrite = sizeof(struct ce2) * count;
    pChar = (char *) pRecompressedEvents;
    while (bytesLeftToWrite) {
	if (bytesLeftToWrite >= DEVICE_BLOCK_SIZE) {
	    writeToFile(fd, pChar, DEVICE_BLOCK_SIZE);
	    bytesLeftToWrite -= DEVICE_BLOCK_SIZE;
	    pChar += DEVICE_BLOCK_SIZE;
	}
	else {
	    writeToFile(fd, pChar, (int) bytesLeftToWrite);
	    break;
	}
    }
}

struct ce2 *recompressFramesEvents(struct ce_crisExpEvt *pExpEvt,
		struct sc_evt_params *PTR_sc_evt_params_structs,
		long max_layer_intensities[][6], int hop_params[],
		float qual_factors[], int eventCount) {
    int iEvent, j, iStack, iGuard, iBlob, iLayer, iXY;
    int CENTER_THRESH[] = { 125, 450, 1000 };
    int GUARD_THRESH[] =  { 125, 450, 1000 };
    short check;

    struct ce2 *pRecompressedEventsStart;
    struct ce2 *pRecompressedEvents;

    pRecompressedEventsStart =
		(struct ce2 *) calloc(eventCount, sizeof(struct ce2));
    pRecompressedEvents = pRecompressedEventsStart;

    for (iEvent = 0; iEvent < eventCount; iEvent++) {
	pRecompressedEvents->sec96 = pExpEvt->EvtData.secs1996;
	pRecompressedEvents->sc_clk = pExpEvt->EvtData.evtno;
	pRecompressedEvents->ev_ofs = pExpEvt->EvtData.evt_offset;
	pRecompressedEvents->ev_len = pExpEvt->EvtData.nbytes;

	if (pExpEvt->EvtData.prim_ctr_ph[1] > pExpEvt->EvtData.prim_ctr_ph[0])
	    pRecompressedEvents->flags = 0x01;
	else pRecompressedEvents->flags = 0x00;
	pRecompressedEvents->flags = pRecompressedEvents->flags |
			((pExpEvt->EvtData.prim_tel_pair << 1) & 0x02) |
			((pExpEvt->EvtData.guard_hi_or   << 2) & 0x04) |
			((pExpEvt->EvtData.got_guard     << 3) & 0x08) |
			((pExpEvt->EvtData.sec_tel_act   << 4) & 0x30) |
			((pExpEvt->EvtData.got_2nd_tel   << 6) & 0x40) |
			((pExpEvt->EvtData.hazard        << 7) & 0x80);

	pRecompressedEvents->bufid = (pExpEvt->EvtData.buffer_id & 0x3F);

	pRecompressedEvents->rng_gd = ((pExpEvt->EvtData.R1 & 0x0F) |
					((pExpEvt->EvtData.G1 << 4) & 0x70));
	pRecompressedEvents->rng_gd2 = ((pExpEvt->EvtData.R2 & 0x0F) |
					((pExpEvt->EvtData.G2 << 4) & 0x70));

	for (iStack = 0; iStack < 10; iStack++) {
	    short centerThreshCmp = 0;
	    for (j = 0; j < 3; j++)
		if (pExpEvt->EvtData.sec_ctr_ph[iStack] >= CENTER_THRESH[j])
			centerThreshCmp++;
	    pRecompressedEvents->stk[iStack] =
		(pExpEvt->EvtData.prim_ctr_ph[iStack] & 0x0FFF) |
		((pExpEvt->EvtData.prim_tag << (12 - iStack)) & 0x1000) |
		((centerThreshCmp << 13) & 0x6000) |
		((pExpEvt->EvtData.sec_tag << (15 - iStack)) & 0x8000);
	}

	for (iGuard = 0; iGuard < 6; iGuard++) {
	    u_short guardThreshCmp = 0, primTagAux;
	    for (j = 0; j < 3; j++) 
		if (pExpEvt->EvtData.sec_guard_ph[iGuard] >= GUARD_THRESH[j])
			guardThreshCmp++;
	    primTagAux = pExpEvt->EvtData.prim_tag;
	    if (iGuard > 2) primTagAux = primTagAux >> (iGuard - 2);
	    else primTagAux = primTagAux << (2 - iGuard);
	    pRecompressedEvents->grd[iGuard] =
		(pExpEvt->EvtData.prim_guard_ph[iGuard] & 0x0FFF) |
		(primTagAux & 0x1000) |
		((guardThreshCmp << 13) & 0x6000) |
		((pExpEvt->EvtData.sec_tag << (5 - iGuard)) & 0x8000);
	}

	if ( (pExpEvt->EvtData.stim != 0x00) &
		((pExpEvt->EvtData.stimdata.ctr_dac_level !=
				pExpEvt->EvtData.stimdata.guard_dac_level) |
		(pExpEvt->EvtData.stimdata.ctr_pulsers != 0x0F) |
		(pExpEvt->EvtData.stimdata.guard_pulsers != 0x0F) |
		(pExpEvt->EvtData.stimdata.ctr_hi_gain != 0x01) |
		(pExpEvt->EvtData.stimdata.guard_hi_gain != 0x01)) )
	    check = 0x4000;
	else check = 0x0000;
	pRecompressedEvents->stim =
		((pExpEvt->EvtData.stimdata.ctr_dac_level & 0xFF) |
		((pExpEvt->EvtData.stimdata.soft_led_pulse << 8) & 0x3F00) |
		check |
		((pExpEvt->EvtData.stim << 15) & 0x8000));

	pRecompressedEvents->npix = compressNpix(pExpEvt->EvtData.npix);

	pRecompressedEvents->soft_par[0] =
		(pExpEvt->EvtData.nblobs & 0x1F) |
		((pExpEvt->EvtData.cam << 6) & 0x40) |
		((pExpEvt->EvtData.centroid_err << 7) & 0x80);
	pRecompressedEvents->soft_par[1] = PTR_sc_evt_params_structs->ntraj;
	    /* nextqual is a float between 0 and 1.  change range 0 - 255 int */
	pRecompressedEvents->soft_par[2] =
			(u_char) (255.0 * PTR_sc_evt_params_structs->nextqual);
	pRecompressedEvents->soft_par[3] = PTR_sc_evt_params_structs->ret_code;
	    /* only one trajectory is returned inside a ce2 struct */
	pRecompressedEvents->soft_par[4] = 1;

	    /* fill max_int fields */
	{
		/* these variables are local to this block only */
	    int i;
	    long *I;
	    short compressed_intensities[6], x_layers, y_layers;

		/* point I at start of array of 6 longs, ordered x1,y1,x2 ... */
	    I = max_layer_intensities[iEvent];
	        /* compress the max intensities for this event */
	    for (i = 0; i < 6; i++) {
		    /* compressInt is supposed to return a 5 bit value */
		compressed_intensities[i] = (short) (compressInt(I[i]) & 0x1f);
	    }

	    x_layers = compressed_intensities[0];
	    x_layers = x_layers | (compressed_intensities[2] <<  5);
	    x_layers = x_layers | (compressed_intensities[4] << 10);
	    x_layers = x_layers & 0x7fff;	/* high order bit is zero */

	    y_layers = compressed_intensities[1];
	    y_layers = y_layers | (compressed_intensities[3] <<  5);
	    y_layers = y_layers | (compressed_intensities[5] << 10);
	    y_layers = y_layers & 0x7fff;	/* high order bit is zero */

	    pRecompressedEvents->max_int[0] = x_layers;
	    pRecompressedEvents->max_int[1] = y_layers;
	}

	    /* fill qual_hop field */
	{
		/* block locals */
	    long temp = 0x00000000;
	    u_char compressed_quality;
	    long hop_info_X1, hop_info_X2, hop_info_X3;
	    long hop_info_Y1, hop_info_Y2, hop_info_Y3;
	    int hop_param = hop_params[iEvent];

	    compressed_quality = (u_char) (qual_factors[iEvent] * 255.0);

		/* keeping it obvious */
	    hop_info_X1 = (long) ((hop_param >> 15) & 07);
	    hop_info_Y1 = (long) ((hop_param >> 12) & 07);
	    hop_info_X2 = (long) ((hop_param >>  9) & 07);
	    hop_info_Y2 = (long) ((hop_param >>  6) & 07);
	    hop_info_X3 = (long) ((hop_param >>  3) & 07);
	    hop_info_Y3 = (long) ((hop_param      ) & 07);

		/* the "&"s should be superfluous, but show field locations */
	    temp = compressed_quality          & 0x000000ff;

	    temp = temp | ((hop_info_X1 <<  8) & 0x00000700);
	    temp = temp | ((hop_info_Y1 << 11) & 0x00003800);
	    temp = temp | ((hop_info_X2 << 14) & 0x0001c000);
	    temp = temp | ((hop_info_Y2 << 17) & 0x000e0000);
	    temp = temp | ((hop_info_X3 << 20) & 0x00700000);
	    temp = temp | ((hop_info_Y3 << 23) & 0x03800000);

		/* zero highest order bits */
	    temp = temp                        & 0x03ffffff;

	    pRecompressedEvents->qual_hop = temp;
	}

	for (iBlob = 0; iBlob < 18; iBlob++) {
	    if (iBlob < pExpEvt->EvtData.nblobs) {
		u_char min;
		int position;

		if (pExpEvt->layer[iBlob] < 0x06) min = pExpEvt->layer[iBlob];
		else min = 0x06;
		pRecompressedEvents->soft_dat[iBlob][0] = (u_char)
			((compressInt(pExpEvt->EvtData.hodo[iBlob].I) & 0x1F) |
			((min << 5) & 0xE0));

		    /* convert blob position (mm) to integer in 50 um */
		    /* units.  round to the nearest value.            */
		position = (int) ((pExpEvt->blobpos[iBlob] / 0.050) + 0.5);
		if (pExpEvt->layer[iBlob] <= 5)
		    pRecompressedEvents->soft_dat[iBlob][1] =
						(u_char) (position & 0xFF);
		else pRecompressedEvents->soft_dat[iBlob][1] = 0x00;

		pRecompressedEvents->soft_dat[iBlob][2] = (u_char)
			((pExpEvt->EvtData.hodo[iBlob].x << 5) & 0x00E0);
		if (pExpEvt->layer[iBlob] <= 5)
		    pRecompressedEvents->soft_dat[iBlob][2] =
				pRecompressedEvents->soft_dat[iBlob][2] |
				((u_char) ((position >> 8) & 0x001F));

		pRecompressedEvents->soft_dat[iBlob][3] = (u_char)
			(((pExpEvt->EvtData.hodo[iBlob].x >> 3) & 0x007F) |
			((pExpEvt->EvtData.hodo[iBlob].y << 7) & 0x0080));

		pRecompressedEvents->soft_dat[iBlob][4] = (u_char)
			((pExpEvt->EvtData.hodo[iBlob].y  >> 1) & 0xFF);
	    }
	    else {
		pRecompressedEvents->soft_dat[iBlob][0] = 0xE0;
		pRecompressedEvents->soft_dat[iBlob][1] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][2] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][3] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][4] = 0x00;
	    }
	}

	if (pExpEvt->goodtraj > 0x00) {
	    pRecompressedEvents->coord[0][0] =
			(((short) ((pExpEvt->x1 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[1][0] =
			(((short) ((pExpEvt->x2 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[2][0] =
			(((short) ((pExpEvt->x3 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[0][1] =
			(((short) ((pExpEvt->y1 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[1][1] =
			(((short) ((pExpEvt->y2 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[2][1] =
			(((short) ((pExpEvt->y3 / 0.0050) + 0.5)) & 0xFFFF);
	}
	else {
	    for (iXY = 0; iXY < 2; iXY++) {
		for (iLayer = 0; iLayer < 3; iLayer++) {
		    pRecompressedEvents->coord[iLayer][iXY] = 0x0000;
		}
	    }
	}

	pRecompressedEvents->blob_id =
		((((int) pExpEvt->nx1)		  & 0x0000001F) |
		((((int) pExpEvt->ny1) << 5)	  & 0x000003E0) |
		((((int) pExpEvt->nx2) << 10)	  & 0x00007C00) |
		((((int) pExpEvt->ny2) << 15)	  & 0x000F8000) |
		((((int) pExpEvt->nx3) << 20)	  & 0x01F00000) |
		((((int) pExpEvt->ny3) << 25)	  & 0x3E000000) |
		((((int) pExpEvt->goodtraj) << 30) & 0xC0000000));

	pExpEvt++;
	PTR_sc_evt_params_structs++;
	pRecompressedEvents++;
    }
    return pRecompressedEventsStart;
}

u_char compressNpix(long npix) {
    static u_char *compressed = NULL;
    static long *uncompressed = NULL;
    int i;
    double d;

    if (compressed == NULL) {
	compressed = (u_char *) calloc(COMPRESSED_ELEMENTS, sizeof(u_char));
	uncompressed = (long *) calloc(COMPRESSED_ELEMENTS, sizeof(long));
	for (i = 0; i < COMPRESSED_ELEMENTS; i++) {
	    compressed[i] = (u_char) i;
	    uncompressed[i] = uncompNpix(compressed[i]);
	}
    }

    d = interpol(compressed, uncompressed, npix) + 0.5;
    return (u_char) d;
}

long uncompNpix(u_char compNpix) {
    long npix = (long) compNpix;
    if (npix  >= 32) {
	long exponent = (npix >> 4) - 2L;
	long fraction = ((npix & 0x0000000F) << 1) | 0x00000021;
	npix = (fraction << exponent) - 1L;
    }
    return npix;
}

double interpol(u_char *compressed, long *uncompressed, long npix) {
    static double *slopes = NULL;
    static const int lastI = COMPRESSED_ELEMENTS - 1;
    int i;

    if (slopes == NULL) {
	slopes = (double *) calloc(COMPRESSED_ELEMENTS, sizeof(double));
	for (i = 0; i < lastI; i++) {
	    slopes[i] = ((double) (compressed[i + 1] - compressed[i])) /
			((double) (uncompressed[i + 1] - uncompressed[i]));
	}
    }

    if ((npix < uncompressed[0]) || (npix > uncompressed[lastI])) {
	fprintf(stderr, "interpol: npix = %ld out of range\n", npix);
	fprintf(stderr, "\tuncompressed[0] = %ld\n", uncompressed[0]);
	fprintf(stderr, "\tuncompressed[lastI] = %ld\n", uncompressed[lastI]);
	exit(1);
    }
    else if (npix == uncompressed[lastI]) return compressed[lastI];
    else {
	for (i = lastI; i >= 0; i--)
	    if (npix >= uncompressed[i])
		return (compressed[i] + ((npix - uncompressed[i]) * slopes[i]));
    }
    return -999.999;	/* should never get here */
}

u_char compressInt(long intensity) {
    int compressedIntensity =
		((int) ((2.0 * log(intensity + 1.0) / log(2.0)) + 0.5)) - 4;
    if (compressedIntensity < 0) return 0x00;
    else return (u_char) compressedIntensity;
}

void processSOFTException(int errorCode) {
    int verbose = VERBOSE_SOFT_EXCEPTIONS_FOR_REJECTED_EVENTS;

    if (!verbose) return;

    switch(errorCode) {
	case 421:
	    if (verbose) fprintf(stderr, "multitraj: not enough blobs\n");
	    break;

	case 422:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 1\n");
	    break;

	case 423:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 2\n");
	    break;

	case 424:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 3\n");
	    break;

	case 425:
	    if (verbose) fprintf(stderr, "multitraj: not all layers\n");
	    break;

	case 426:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 1\n");
	    break;

	case 427:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 2\n");
	    break;

	case 428:
	    if (verbose) fprintf(stderr, "multitraj: no Y fit 1\n");
	    break;

	case 429:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 2\n");
	    break;

	case 430:
	    if (verbose) fprintf(stderr, "multitraj: no fit\n");
	    break;

	case 1000:
	    if (verbose) fprintf(stderr, "multitraj: negative detector rad\n");
	    break;

	case 1001:
	    if (verbose) fprintf(stderr, "multitraj: bad stack value\n");
	    break;

	case 1002:
	    if (verbose) fprintf(stderr, "multitraj: trajectory off E1 det\n");
	    break;
    }
}

int get_command_line_args(int argc, char *argv[], char user[],
				char institution[], char output_directory[]) {
    int i;

    for (i = 0; i < argc; i++) {
	if (*(argv[i]) == '-') {
	    char flag_char = *(argv[i] + 1);
	    switch(flag_char) {
		int j;

		case 'd':	/* output directory */
		    strcpy(output_directory, argv[i + 1]);
		    for (j = i + 2; j < argc; j++) { argv[j - 2] = argv[j]; }
		    i--;
		    argc -= 2;
		    break;

		case 'i':	/* institution = (caltech|gsfc|washu) */
		    strcpy(institution, argv[i + 1]);
		    for (j = i + 2; j < argc; j++) { argv[j - 2] = argv[j]; }
		    i--;
		    argc -= 2;
		    break;

		case 'u':	/* user */
		    strcpy(user, argv[i + 1]);
		    for (j = i + 2; j < argc; j++) { argv[j - 2] = argv[j]; }
		    i--;
		    argc -= 2;
		    break;

		default:
		    break;
	    }
	}
    }

    return argc;
}

void print_usage() {
    fprintf(stderr, "Usage: generateCrisEvents_ce2 ");
    fprintf(stderr, "startDayOf1996 endDayOf1996 \\\n");
    fprintf(stderr, "\t-i (caltech|washu|gsfc) ");
    fprintf(stderr, "[-d output_directory] ");
    fprintf(stderr, "[-u user]\n\n");
}

void get_version_information_from_version_files(char directory[],
			char *version_file_names[],
			struct output_file_header_info *PTR_header_info) {
    FILE *file_handle;
    char file_name[80];
    char buffer[81];
    char *PTR_char = NULL;
    int i;

    for (i = 0; i < 5; i++) {
	memset(buffer, '\0', 81);
	sprintf(file_name, "%s/%s", directory, version_file_names[i]);
	/* printf("%s\n", file_name); */
	file_handle = fopen(file_name, "r");
	fgets(buffer, 80, file_handle);
	fclose(file_handle);
	    /* truncate before any new lines present */
	PTR_char = strchr(buffer, '\n');
	if (PTR_char != NULL) { *PTR_char = '\0'; }
	PTR_char = strchr(buffer, '\r');
	if (PTR_char != NULL) { *PTR_char = '\0'; }

	switch (i) {
	    case 0:
		sprintf(PTR_header_info->level1_version,
					"level1: %-72s", buffer);
		break;
	    case 1:
		sprintf(PTR_header_info->level1_1_version,
					"level1_1: %-70s", buffer);
		break;
	    case 2:
		sprintf(PTR_header_info->mk_ce2_version,
					"mk_ce2: %-72s", buffer);
		break;
	    case 3:
		sprintf(PTR_header_info->softlib_version,
					"softlib: %-71s", buffer);
		break;
	    case 4:
		sprintf(PTR_header_info->compile_date,
					"compiled: %-70s", buffer);
		break;
	}
    }

/*
    printf("%s", PTR_header_info->level1_version);
    printf("%s", PTR_header_info->level1_1_version);
    printf("%s", PTR_header_info->mk_ce2_version);
    printf("%s", PTR_header_info->softlib_version);
    printf("%s", PTR_header_info->compile_date);
*/
}

void write_ce2_file_header(int fd,
			struct output_file_header_info *PTR_header_info) {
    char buffer[81];
    int i, endian_pattern_repetitions;

    endian_pattern_repetitions = 80 / sizeof(int);
    for (i = 0; i < endian_pattern_repetitions; i++) {
	char *PTR_char = buffer + (i * sizeof(int));

	    /* this copies the contents of an int as if it were a char[] */
	memcpy(PTR_char, (char *) (&(PTR_header_info->endian_pattern)),
								sizeof(int));
    }
    writeToFile(fd, buffer, 80);

    sprintf(buffer, "institution: %-67s", PTR_header_info->institution);
    writeToFile(fd, buffer, 80);

    writeToFile(fd, PTR_header_info->compile_date, 80);

    sprintf(buffer, "user: %-74s", PTR_header_info->user);
    writeToFile(fd, buffer, 80);

    sprintf(buffer, "run: %-75s", PTR_header_info->run_date);
    writeToFile(fd, buffer, 80);

    writeToFile(fd, PTR_header_info->mk_ce2_version, 80);
    writeToFile(fd, PTR_header_info->softlib_version, 80);
    writeToFile(fd, PTR_header_info->level1_version, 80);
    writeToFile(fd, PTR_header_info->level1_1_version, 80);

    sprintf(buffer, "cris archive: %-66s",
			PTR_header_info->data_archive_version);
    writeToFile(fd, buffer, 80);
}
