/* ce2 file header information */

struct output_file_header_info {
    int endian_pattern;
    char institution[81];
    char compile_date[81];
    char user[81];
    char run_date[81];
    char mk_ce2_version[81];
    char softlib_version[81];
    char level1_version[81];
    char level1_1_version[81];
    char data_archive_version[81];
};

/* ce2.h - mew's recompressed cris event format */

struct ce2 {
    long sec96;
    long sc_clk;
    u_char ev_ofs;
    u_char ev_len;
    u_char flags;
    u_char bufid;
    u_char rng_gd;
    u_char rng_gd2;
    short stk[10];
    short grd[6];
    short stim;
    u_char npix;
 
/*  u_char soft_par;            // old form in ce1.  changed */
    u_char soft_par[5];         /* new form */
 
    short max_int[2];           /* new to ce2 */
 
    u_char soft_dat[18][5];
    short coord[3][2];
    long blob_id;
 
/*  u_char status[4];           // old from ce1.  removed from ce2 */
 
    long qual_hop;              /* new to ce2 */
};

/*
    * ----------------------------------------------------------------
    * Documentation of output structure format and contents.
    *
    * SEC96   - Time of the start of the event cycle in which this
    *           event was collected, in seconds since the start of 1996.
    *           Note that all events collected in the same 256 second
    *           instrument cycle have identical values of this quantity.
    * SC_CLK  - Spacecraft clock (raw counter) at the time this event
    *           was readout.  In seconds.
    * EV_OFS  - Offset of the start of this event from the start of the
    *           CRIS data block for the minor frame in which the event
    *           was readout.  The combination of "SC_CLK" and "EV_OFS"
    *           constitute a "unique event ID" which can be used for
    *           referring to a specific event.
    * EV_LEN  - Length of this event in the CRIS telemetry stream, in bytes.
    * FLAGS   - Collection of various indicator bits, as follows:
    *           Bit 0 - telescope in primary telescope pair which
    *                   has the largert E1 pulse height (in channels)
    *                   (0 = A or C, 1 = B or D)
    *           Bit 1 - Primary telescope pair (0=AB, 1=CD)
    *           Bit 2 - logical OR of low level guards
    *           Bit 3 - flag to indicate whether guards were transmitted
    *                   in the event (0=not transmitted, 1=transmitted)
    *           Bits 4 & 5
    *                 - Second telescope activity. Two bits indicate
    *                   whether any centers and/or guards were hit
    *                   in the secondary telescope (Need to figure out
    *                   the details of how this information is encoded
    *                   by checking with Rick Cook and the flight firmware
    *                   listing and/or by examining the events which have
    *                   the possible combinations of these bits.)
    *           Bit 6 - Indicates whether pulse heights from secondary
    *                   telescope were included in this event (0=not
    *                   included, 1=included)
    *           Bit 7 - Hazard flag for this event (0=no hazard, 1=hazard)
    * BUFID   - Bits 0-5
    *                 - ID number of readout buffer (and priority)
    *           Bits 6-7
    *                 - unused -- should always be 0
    * RNG_GD  - Bits 0-3
    *                 - Range in primary telescope (number of the stop
    *                   detector.  (In the raw telemetry "2" is subracted
    *                   from this value before transmission, so that only
    *                   3 bits will be needed for this value.  Here the
    *                   value is the actual range without the -2.)
    *         - Bits 4-6
    *                 - Number of guards hit in primary telescope
    * RNG_GD2 - Bits 0-3
    *                 - Range in secondary telescope (number of the stop
    *                   detector.  (In the raw telemetry "2" is subracted
    *                   from this value before transmission, so that only
    *                   3 bits will be needed for this value.  Here the
    *                   value is the actual range without the -2.)
    *         - Bits 4-6
    *                 - Number of guards hit in secondary telescope
    * STK     - Array of center pulse heights from the primary telescope,
    *           in the order E1A, E1B, E2AB, E3AB, ..., E9AB (or the
    *           equivalent for CD telescope), together with tab bits from
    *           both the primary and secondary telescopes and a rough
    *           indication of the pulse heights in the secondary telescope
    *           (if any).  Bits are used as follows:
    *         - Bits 0-11
    *           - Pulse height (channels) from this detector in
    *             the primary telescope.
    *         - Bits 12-13
    *           - Encoded value giving a coarse indication of the pulse
    *             height (if any) in this detector in the secondary
    *             telescope.  Encoding is TBD, but might be something
    *             like 0 => <100, 1 => 100-299, 2 => 300-999, 3 => >999
    *         - Bit 14
    *           - Tag bit for this detector in secondary telescope.
    *         - Bit 15
    *           - Tag bit for this detector in primary telescope.
    * GRD     - Array of guard pulse heights from the primary telescope,
    *           in the order G2AB, G3AB, G4AB, G5AB, G6AB, G7AB (or the
    *           equivalent for CD telescope), together with tab bits from
    *           both the primary and secondary telescopes and a rough
    *           indication of the pulse heights in the secondary telescope
    *           (if any).  Fields are analogous to those used for the "STK"
    *           array defined above.  Bits are used as follows:
    *         - Bits 0-11
    *           - Pulse height (channels) from this detector in
    *             the primary telescope.
    *         - Bits 12-13
    *           - Encoded value giving a coarse indication of the pulse
    *             height (if any) in this detector in the secondary
    *             telescope.  Encoding is TBD, but might be something
    *             like 0 => <100, 1 => 100-299, 2 => 300-999, 3 => >999
    *         - Bit 14
    *           - Tag bit for this detector in secondary telescope.
    *         - Bit 15
    *           - Tag bit for this detector in primary telescope.
    * STIM    - Collection of quantities related to the "stim" events
    *           generated by the on-board pulsers.
    *           Bits 0-7
    *           - DAC setting used to set the stim pulse height.
    *           Bits 8-13
    *           - Length of the LED pulse used to stimulate SOFT.
    *             (What are the time units?)
    *           Bit 14
    *           - Non-standard stim conditions.  Bit is normally 0.  It
    *             is set to 1 if: (a) the various DAC values reported
    *             in the raw event are not all equal or (b) the flags
    *             indicating which guard or center pulsers are active
    *             are not all in their normal states or (c) the bits
    *             indicating the DAC gain settings are not all in their
    *             normal states.
    *           Bit 15 -
    *           - Stim event.  Bit is 1 if this is a stim event, 0 otherwise.
    *             If this is not a stim event all other bits in this byte
    *             are meaningless and should be forced to 0.
    * NPIX    - Number of pixels above threshold in the SOFT image.
    *           This is encoded as an 8-bit floating point value.
    * SOFT_PAR - Parameters related to SOFT
    *            byte 0: bits 0-4: number of clusters
    *                    bit    5: unused (set to 0)
    *                    bit    6: Active camera, a guess (0=A, 1=B)
    *                    bit    7: SOFT camera error (0=no error, 1=error)
    *            byte 1: number of trajectories found
    *            byte 2: quality of next-best trajectory
    *            byte 3: return code from sc_cris_traj
    *            byte 4: number of trajectories returned = 1
    * max_int  - maximum intensity in each layer
    *            bytes 0 - 1: bits  0 -  4: maximum intensity in layer X1
    *                         bits  5 -  9: maximum intensity in layer X2
    *                         bits 10 - 14: maximum intensity in layer X3
    *                         bit       15: unused (set to 0)
    *            bytes 2 - 3: bits  0 -  4: maximum intensity in layer Y1
    *                         bits  5 -  9: maximum intensity in layer Y2
    *                         bits 10 - 14: maximum intensity in layer Y3
    *                         bit       15: unused (set to 0)
    * SOFT_DAT   - BYTARR(5,18)
    *              Array containing 5 bytes for each possible SOFT blob
    *              in the raw data.  2nd index is the blob number.
    *              The blobs appear in the same order as they are
    *              found in the raw event data.
    *              The 5 bytes contain the following quantities:
    *              Byte 0, bits 0-4
    *              - Blob intensity (encoded as in the raw data)
    *              Byte 0, bits 5-7
    *              - Blob layer (0=Y3, 1=X3, 2=Y2, 3=X2, 4=Y3, 5=X3)
    *                A value of 7 is used if the blob centroid does
    *                not fall in a valid layer.
    *              Byte 1 all bits & Byte 2 bits 0-4
    *              - Rough coordinate value (no correction for the
    *                value of the orthogonal coordinate).  13-bit
    *                integer value in units of 50 microns with possible
    *                offset TBD.
    *                Byte 1 contains the low order 8 bits of this quantity.
    *              Byte 2 bits 5-7 & Byte 3 bits 0-6
    *              - x centroid of blob (from raw event)
    *                Byte 2 contains the low order 3 bits of this 10-bit
    *                quantity.
    *              Byte 3 bit 7 & Byte 4 all bits
    *              - y centroid of blob (from raw event)
    *                Byte 3 contains the low order 1 bit of this 9-bit
    *                quantity.
    * COORDS     - INTARR(6)
    *              Selected coordinates in each of the 6 SOFT layers in
    *              the order X1, Y1, X2, Y2, X3, Y3.
    *              Each coordinate is a 16-bit signed in units of 5 microns
    *              with a possible offset TBD.
    * BLOB_IDS   - Collection of blob numbers used to obtain the trajectory
    *              coordinates given as COORDS, above.  5 bits of the 32-bit
    *              word are allocated to each layer, as follows:
    *              X1: bits  0- 4;  Y1: bits  5- 9
    *              X2: bits 10-14;  Y2: bits 15-19
    *              X3: bits 20-24;  Y3: bits 25-29
    *              Bits 30-31:
    *                  0 indicates no trajectory was found
    *                  2 indicates good trajectory
    *                  (two bits were retainded for consistency with
    *                      previous programs)
    * qual_hop - quality of this trajectory and hop information
    *            bits  0 -  7: quality value of this trajectory
    *            bits  8 - 10: hop info for coordinate X1
    *            bits 11 - 13: hop info for coordinate Y1
    *            bits 14 - 16: hop info for coordinate X2
    *            bits 17 - 19: hop info for coordinate Y2
    *            bits 20 - 22: hop info for coordinate X3
    *            bits 23 - 25: hop info for coordinate Y3
    *            bits 26 - 31: unused (set to 0)
    * ---------------------------------------------------------
*/
